import numpy as np
import sys, os
import utilities, simcalc


def normArray(handle, d, coeffs, src, dest, start, end):
	if start == None: start = 0
	if end == None: end = handle.lengthInFrames
	if d.seg:
		if handle.typed == 'target':
			for fidx in range(start, end):
				if handle.data[src][d.name][fidx] == None: continue
				handle.data[dest][d.name][fidx] = (handle.data[src][d.name][fidx]-coeffs[0])/coeffs[1]
		else:
			handle.data[dest][d.name][(0, None)] = (handle.data[src][d.name][(0, None)]-coeffs[0])/coeffs[1]
	else:
		handle.data[dest][d.name][start:end] = (handle.data[src][d.name][start:end]-coeffs[0])/coeffs[1]

			




def normalizeDescriptors(whichHandles, whichAnals, dataIn, dataOut, normType, normCoefDict, s, e):
	for handle in whichHandles:
		for d in whichAnals:
			if not d.seg:
				if s == None and e == None:
					#print d.name, handle.data[dataIn][d.name], normType, normCoefDict
					handle.data[dataOut][d.name] = normDataArray(d.name, handle.data[dataIn][d.name][:], normType, normCoefDict)
				else:
					handle.data[dataOut][d.name][s:e] = normDataArray(d.name, handle.data[dataIn][d.name][s:e], normType, normCoefDict)
			else:
				handle.segdata.setNormCoeffs(normCoefDict[d.name][0], normCoefDict[d.name][1])


									

def normDataArray(anal, dataList, normType, normCoefDict):
	output = []
	# BY MIN MAX
	if normType == 'minmax':
		if type(dataList) == list:
			for val in dataList:
				if val == None: output.append(None)
				else: output.append(  (val-normCoefDict[anal][0])/(normCoefDict[anal][1]-normCoefDict[anal][0])  )
		else: # its a numpy array
			output = (dataList-normCoefDict[anal][0])/(normCoefDict[anal][1]-normCoefDict[anal][0])
			
	# BY STANDARD DEVIATION
	else:
		if type(dataList) == list:
			for val in dataList:
				if val == None: output.append(None)
				else: output.append(  (val-normCoefDict[anal][0])/(normCoefDict[anal][1])/ops.NORMALIZE_SCALE_STD_DEV  )
		else: # its a numpy array
			output = (dataList-normCoefDict[anal][0])/(normCoefDict[anal][1])/ops.NORMALIZE_SCALE_STD_DEV
	#print output
	return output



def makeTgtClusterNormCoefDict(timeInFrames, anal, clusterNormCoeffs, targetNormCoeffs, sourceNormCoeffs, tgtClusterAssignments, analMetrics):
	clusterNumb = tgtClusterAssignments[timeInFrames]
	if clusterNumb == None : clusterNumb = 0
	return {'target': clusterNormCoeffs['target'][clusterNumb][anal], 'corpus': clusterNormCoeffs['corpus'][clusterNumb][anal], 'allCorpus': sourceNormCoeffs[anal], 'allTarget': targetNormCoeffs[anal], 'userCoeff': analMetrics.metric[anal]['normMethod']}





def getNormCoefs(whichHandles, d, start, end, normMethod):
	hugeAnalList = [] # pointers only, so wont FUCK me too hard...
	for handle in whichHandles: # puts all data of one anal method into one huge list
		if handle.typed == 'target':
			# we are only taking segment data...
			if ops.NORMALIZE_TARGET_SEGMENTS_ONLY and handle.typed == 'target' and start == None and end == None: 
				if not d.seg:
					for st, ed in handle.segFrames:
						hugeAnalList.extend(handle.data['raw'][d.name][st:ed]) # add a segment slice
				else:
						hugeAnalList.extend(handle.segdata.rawvalues(d.name)) # add a segment slice
			elif start != None and end != None: # only taking a single target segment
				# broken seg descriptor
				hugeAnalList.extend(handle.data['raw'][d.name][start:end]) # add a segment slice
			else:
				if not d.seg:
					hugeAnalList.extend(handle.data['raw'][d.name]) # add the whole thing		
				else:
					hugeAnalList.extend(handle.segdata.rawvalues(d.name)) # add the whole thing
						
		elif handle.typed == 'source': # we only are taking the eff duration..
			if not d.seg:

				hugeAnalList.extend(handle.data['raw'][d.name][0:handle.segdata.getraw('effDur-seg', 0, None)]) # add a segment slice
			else:
				hugeAnalList.append(handle.segdata.getraw(d.name, 0, handle.segdata.getraw('effDur-seg', 0, None))) # add a segment slice
	#print "1", whichHandles, anal, start, end
	if normMethod in ['stddev', 'cluster']:
		# scale based on std dev and mean
		stdy = np.std(hugeAnalList)
		if stdy == 0: stdy = 0.01 # prevent divisions by zero later on.
		return np.mean(hugeAnalList), stdy # 3.5 via norbert... make outliers less extreme
	elif normMethod == 'minmax':
		# scale based on min max
		return min(hugeAnalList), max(hugeAnalList)



def normalizeCoefs(descriptor_list, tgt, ag, p, normMethod):
	#print descriptor_list, tgt, ag, p
	p.post("Normalizing based upon Standard Deviation...", 3)
	outNormalizeDictTgt = {} # a dict holding d.name min/max as lists, d.name d.norm is key
	outNormalizeDictSrc = {} # a dict holding d.name min/max as lists, d.name d.norm is key
	for idx, d in enumerate(descriptor_list):
		[tgtMean, tgtStd] = getNormCoefs([tgt], d, None, None, normMethod)
		[srcMean, srcStd] = getNormCoefs(ag.sourceHandles, d, None, None, normMethod)
		[allMean, allStd] = getNormCoefs(ag.allHandles, d, None, None, normMethod)

		if d.norm > 0:
			ag.normPrintData[d.name] = {'sign': '+'}
		else:
			ag.normPrintData[d.name] = {'sign': '-'}

		if abs(d.norm) > 0 and abs(d.norm) <= 1: # norm together coeffs
			outNormalizeDictTgt[d.name] = [allMean, allStd]
			outNormalizeDictSrc[d.name] = [allMean, allStd]
			ag.normPrintData[d.name]['norm'] = 'together'
			ag.normPrintData[d.name]['percent'] = int(abs(d.norm)*100)

		elif abs(d.norm) > 1 and abs(d.norm) < 2: # norm mixture coeffs
			outNormalizeDictTgt[d.name] = [utilities.interpolateVal(d.norm-1, allMean, tgtMean), utilities.interpolateVal(d.norm-1, allStd, tgtStd)]
			outNormalizeDictSrc[d.name] = [utilities.interpolateVal(d.norm-1, allMean, srcMean), utilities.interpolateVal(d.norm-1, allStd, srcStd)]
			ag.normPrintData[d.name]['norm'] = 'between'
			ag.normPrintData[d.name]['percent'] = int((abs(d.norm)-1)*100)

		elif abs(d.norm) >= 2: # norm separate coeffs
			outNormalizeDictTgt[d.name] = [tgtMean, tgtStd]
			outNormalizeDictSrc[d.name] = [srcMean, srcStd]
			ag.normPrintData[d.name]['norm'] = 'separate'
			ag.normPrintData[d.name]['percent'] = int((abs(d.norm)/2.0)*100)

#		# now modify ops.TARGET_DESCRIPTORS_MOD to explode/compress/invert features per the user...
#		if abs(d.norm) < 1: # compress from together
#			ops.TARGET_DESCRIPTORS_MOD.append([d.name, d.norm, '*'])
#		elif abs(d.norm) > 2: # explode from sep
#			ops.TARGET_DESCRIPTORS_MOD.append([d.name, d.norm/2.0, '*'])
#		elif d.norm == -1:
#			ops.TARGET_DESCRIPTORS_MOD.append([d.name, d.norm, '*'])
#		elif d.norm == -2:
#			ops.TARGET_DESCRIPTORS_MOD.append([d.name, d.norm/2.0, '*'])
#	print outNormalizeDictTgt
#	print outNormalizeDictSrc
	return outNormalizeDictTgt, outNormalizeDictSrc


def getDescriptorInfo(anal, analMetrics):
	return (anal.find('-seg') != -1, analMetrics.ifDescriptorHasTimeVaryingParent(anal), analMetrics.getTimeVaryingParent(anal)[0], analMetrics.metric[anal]['distanceWeight'], analMetrics.metric[anal]['normMethod'])





def targetTimbreReassignment(descriptorData, tgt, sourceHandles, dataSlot, agData, WEIGHT_AVERAGES_WITH_AMPLITUDE=False, PLOT_REMAP=False):
	remapDescripts = []
	notRemapDescripts = []
	noNotRemap = ['power', 'rms', 'power-seg', 'rms-seg']
	needToReevalTgtSegments  = []
	print "Target Timbre Remapping"
	
	
	for d in descriptorData.master_list:
		if d.weight == 0: continue
		if d.origin == 'TARGET_ONSET_DESCRIPTORS': continue
		if d.name in noNotRemap:
			notRemapDescripts.append(d)
			continue
		if d.seg:
			if len(d.parents) == 0:
				notRemapDescripts.append(d)
			else:
				needToReevalTgtSegments.append(d)
				pd = descriptorData.getDescriptorObjByName(d.parents[0])
				remapDescripts.append(pd)
		else:
			remapDescripts.append(d)
	
	checksum = utilities.getTargetCorpusTimbreRemapChecksum(tgt, sourceHandles, remapDescripts)

	# look if its already been computed and stored on disk
	if not agData.hasData(checksum): # needs to be computed
		print "\tauto-associative multivariate regression trees courtesy of Dan Stowell"
		# do the analysis
		# get number of md points for tgt and corpus collections
		segIdxs = []
		for s in tgt.segFrames: segIdxs.extend(range(s[0],s[1]))
		cpsPoints = 0
		for corpusIdx, handle in enumerate(sourceHandles): 
			cpsPoints += h.segdata.getraw('effDur-seg', 0, None)
		# fill TARGET n-D array
		ddim = len(remapDescripts)
		tgtAllCoors = np.empty((tgt.lengthInFrames, ddim+1))
		tgtSegCoors = np.empty((len(segIdxs), ddim+1))
		sf = 0
		
		for f in range(tgt.lengthInFrames):
			tgtDescript = np.empty(ddim+1)
			# fill tgt datum
			for didx, d in enumerate(remapDescripts):
				tgtDescript[didx] = tgt.data[dataSlot][d.name][f]
			tgtDescript[-1] = f
			tgtAllCoors[f] = tgtDescript
			if f in segIdxs:
				tgtSegCoors[sf] = tgtDescript
				sf += 1
		# fill CORPUS n-D array
		cpsCoors = np.empty((cpsPoints, ddim+1)) # last dim holds frame index as float == sourceIdx.frame
		cpsFrameToPower = {}
		pnt = 0
		for corpusIdx, h in enumerate(sourceHandles): 
			for f in range(0, h.segdata.getraw('effDur-seg', 0, None)):
				for didx, d in enumerate(remapDescripts):
					if d.seg: cpsCoors[pnt][didx] = h.data[dataSlot][d.name][(0, None)]
					else: cpsCoors[pnt][didx] = h.data[dataSlot][d.name][f]
				# associate this frame with its power
				Idx = corpusIdx+(float(f)/100.)
				cpsCoors[pnt][-1] = Idx
				cpsFrameToPower[Idx] = h.data['raw']['power'][f]
				pnt += 1
		# do regression tree analysis
		import Xamrt
		p = Xamrt.Xamrt(tgtSegCoors, cpsCoors, ddim, sizethresh=0, maxdepth=9999)
		# create output format, as lists for json storage
		tgtDataRemap = {}
		for d in remapDescripts:
			tgtDataRemap[d.name] = np.empty(tgt.lengthInFrames)
		tgtNodeTracker = {}
		# map corpus leafs onto the target
		for f in range(tgt.lengthInFrames):
			# classfy in its tree
			tgtNode = p.classify(tgtAllCoors[f], 0)
			if not tgtNodeTracker.has_key(tgtNode): tgtNodeTracker[tgtNode] = 1
			else: tgtNodeTracker[tgtNode] += 1
			# get associated arrays on leafs
			tgtArrays, cpsArrays  = p.getleaf(tgtNode)
			#print f, len(tgtArrays), len(cpsArrays)
			if WEIGHT_AVERAGES_WITH_AMPLITUDE:
				# weight corpus frame sums with each frame's power-seg
				averageCps = np.zeros(ddim)
				sumweights = 0
				for ar in cpsArrays:
					power = cpsFrameToPower[ar[-1]] # get this frames power from dict..
					averageCps += ar[:-1]*power
					sumweights += power
				averageCps /= sumweights
			else:
				averageCps = np.average(cpsArrays, axis=0)
			# add new values to tgtDataRemap dict
			for didx, d in enumerate(remapDescripts):
				tgtDataRemap[d.name][f] = averageCps[didx]
			
			jsonSerial = {}
			for key, val in tgtDataRemap.iteritems():
				jsonSerial[key] = list(val)
			# add to storage file
			agData.commitData(checksum, jsonSerial)	
	else:
		# read it from disk!
		print "\tloading analysis - auto-associative multivariate regression trees courtesy of Dan Stowell"
		tgtDataRemap = agData.getData(checksum)

	if PLOT_REMAP:
		numSegs = 8
		print "\tplotting original and remapped target descriptors..."
		stop = min(tgt.lengthInFrames, 230)
		sl = slice(0, stop) # a slice of "numSegs" segments in length
		for didx, (descript, isseg) in enumerate(remapDescripts):
			utilities.plot('/Users/ben/Desktop/remap/%s.png'%(descript), [tgt.data[dataSlot][descript][sl], tgtDataRemap[descript][sl]], ['norm', 'remap'])

	# put the new data into the target descriptor arrays :)
	for descript in tgtDataRemap:
		tgt.data[dataSlot][descript] = np.array(tgtDataRemap[descript]) # the y got converted to lits for json

	return needToReevalTgtSegments, notRemapDescripts








def clusterMeanStd(coors, realIdxs):
	import simcalc
	nSamples = len(coors)
	dim = coors.shape[1]
	meanPoint = np.mean(coors, axis=0)
	std = np.std(coors, axis=0)
	mins = np.amin(coors, axis=0)
	maxs = np.amax(coors, axis=0)
	diff = maxs-mins
	centroid = (diff/2.0)+mins
	# get closest point
	distances = []
	for i, coor in enumerate(coors):
		geometricDistance = np.sum(np.power((coor-meanPoint), 2))
		distances.append((geometricDistance, i))
	distances.sort()
	winnerIdx = distances[0][1]
	maxDist = distances[-1][0]
	if maxDist == 0:
		density = 1
	else:
		density = nSamples/maxDist
	return {'bestPoint': coors[winnerIdx], 'avg': meanPoint, 'std': std, 'bestIdx': realIdxs[winnerIdx], 'clustSize': len(coors), 'minmaxDist': (distances[0][0], distances[-1][0]), 'clustDensity': density, 'mins': mins, 'maxs': maxs, 'dev': diff, 'centroid': centroid}





def normalizeClusterCoefs(numberClusters, normMetrics, tgt, sourceHandles, tgtAssign, cpsAssign):
	IF_STD_IS_ZERO = 0.5
	clusterNormCoeffs = {'target': {}, 'corpus': {}}
	clusterNormMinMax = {'target': {}, 'corpus': {}}
	for i in range(numberClusters):
		clusterNormCoeffs['target'][i] = {}
		clusterNormMinMax['target'][i] = {}
		for anal in normMetrics:
			tmpHugeAnalList = []
			for segIdx, (st, ed) in enumerate(tgt.segFrames):
				if tgtAssign[segIdx] !=  i: continue
				tmpHugeAnalList.append(tgt.data['raw'][anal][st:ed])
			longy = utilities.chain(tmpHugeAnalList)
			hugeAnalListNoNones = []
			for item in longy: # get rid of None entries in segment anals
				if item != None: hugeAnalListNoNones.append(item)	
			standardD = np.std(hugeAnalListNoNones)
			if standardD == 0: standardD = IF_STD_IS_ZERO
			clusterNormCoeffs['target'][i][anal] = (np.mean(hugeAnalListNoNones), standardD)
			clusterNormMinMax['target'][i][anal] = (np.min(hugeAnalListNoNones), np.max(hugeAnalListNoNones))
	
		clusterNormCoeffs['corpus'][i] = {}
		clusterNormMinMax['corpus'][i] = {}
		for anal in normMetrics:
			tmpHugeAnalList = []
			for segIdx, handle in enumerate(sourceHandles):
				if cpsAssign[segIdx] != i: continue
				tmpHugeAnalList.append(handle.data['raw'][anal][0:handle.segdata.getraw('effDur-seg', 0, None)])
			longy = utilities.chain(tmpHugeAnalList)
			hugeAnalListNoNones = []
			for item in longy: # get rid of None entries in segment anals
				if item != None: hugeAnalListNoNones.append(item)
			standardD = np.std(hugeAnalListNoNones)
			if standardD == 0: standardD = IF_STD_IS_ZERO
			clusterNormCoeffs['corpus'][i][anal] = (np.mean(hugeAnalListNoNones), standardD)
			clusterNormMinMax['corpus'][i][anal] = (np.min(hugeAnalListNoNones), np.max(hugeAnalListNoNones))
	return clusterNormCoeffs, clusterNormMinMax





def normalizeAccordingToClusterMapping(tgt, handles, assign, clusterMapping, clusterNormMinMax, analMetrics, ag):
	for segIdx, handle in enumerate(handles):
		cpsClusterId = assign[segIdx]
		tgtClusterId = [i for i,x in enumerate(clusterMapping) if x == assign[segIdx]][0] # backwards fucking mapping
		for anal in analMetrics.norm:
			if analMetrics.metric[anal]['normMethod'] == 2:
				normInClusterId = np.array(normDataArray(anal, handle.data['raw'][anal][0:], 'stddev', ag.clusterNormCoeffs['corpus'][cpsClusterId]))
				normInClusterId = handle.data['raw'][anal][0:]
				# recast raw data as fake target values
				handle.data['raw'][anal] = (normInClusterId*ag.clusterNormCoeffs['target'][tgtClusterId][anal][1])+ag.clusterNormCoeffs['target'][tgtClusterId][anal][0]
				handle.data['norm'][anal] = np.array(normDataArray(anal, handle.data['raw'][anal][0:], 'stddev', ag.targetNormCoeffs))
			else:
				handle.data['norm'][anal] = np.array(normDataArray(anal, handle.data['raw'][anal][0:], 'stddev', ag.targetNormCoeffs))





def unnormalizeAccordingToClusterMappingMinMax(tgt, handles, assign, clusterMapping, clusterMinMax, clustStddev, clusterMetricNames, ag, method='minmax'):
	# get mins/maxs based on anal methods (from coordinates, kinda fucked up)
	# get avgs/stds based on anal methods (from coordinates, kinda fucked up)
	minmax = {}
	stddev = {}
	for name in clusterMinMax:
		minmax[name] = []
		stddev[name] = []
		for cidx, (mins, maxs) in enumerate(clusterMinMax[name]):
			minmax[name].append({})
			for aidx in range(len(mins)):
				miny = mins[aidx]
				maxy = maxs[aidx]
				if miny == maxy : maxy += 0.1 # add some to max for dimensionallity
				minmax[name][cidx][clusterMetricNames[aidx]] = (miny, maxy)
		for cidx, (avgs, stds) in enumerate(clustStddev[name]):
			stddev[name].append({})
			for aidx in range(len(avgs)):
				#$stddev[name][cidx][clusterMetricNames[aidx]] = (avgs[aidx], stds[aidx])
				avg = avgs[aidx]
				std = stds[aidx]
				if std == 0 : std += 0.1 # add some to std for dimensionallity
				stddev[name][cidx][clusterMetricNames[aidx]] = (avg, std)
	print "\tNormalize according to cluster's "+method
	for segIdx, handle in enumerate(handles):
		cpsClusterId = assign[segIdx]
		tgtClusterId = [i for i,x in enumerate(clusterMapping) if x == assign[segIdx]][0] # backwards fucking mapping
		for aidx, anal in enumerate(clusterMetricNames):
			if method == 'minmax':
				(cminy,cmaxy) = minmax['corpus'][cpsClusterId][anal]
				(tminy,tmaxy) = minmax['target'][tgtClusterId][anal]
				normInCluster = (handle.data['raw'][anal]-cminy)/(cmaxy-cminy)
				handle.data['raw'][anal] = (normInCluster*(tmaxy-tminy))+tminy
			elif method == 'stddev':
				(cavg,cstd) = stddev['corpus'][cpsClusterId][anal]
				(tavg,tstd) = stddev['target'][tgtClusterId][anal]
				normInCluster = (handle.data['raw'][anal]-cavg)/cstd
				handle.data['raw'][anal] = (normInCluster*tstd)+tavg
			

					


def normalizeByCluster(analMetrics, tgt, ag, agData):
	MAKE_PLOTS = True
	clusterMetricNames = []
	clusterMetricWeights = []
	for anal in analMetrics.norm:
		if analMetrics.metric[anal]['normMethod'] != 2: continue # only metrics = 2 norm coeff
		if anal.find('-seg') == -1: clusterMetricNames.append(anal+'-seg')
		else: clusterMetricNames.append(anal)
		clusterMetricWeights.append( analMetrics.metric[anal]['distanceWeight'] )
	# get unique clustering checksum
	checksum = utilities.getTargetCorpusChecksum(ops, tgt, ag.sourceHandles, clusterMetricNames)
	# look if its already been computed and stored on disk
	clusterData = agData.getData("cluster", checksum)
	if clusterData == None: # then do the analysis
		print "Computing Clustering Analysis ..."
		clusterNodes, tgtAssign, cpsAssign, clustMinmax, clustStddev = clusterAnal(clusterMetricNames, np.array(clusterMetricWeights), tgt, ag.sourceHandles, ops.DESCRIPTOR_HOP_SIZE_SEC, NORMALIZE_NUMBER_CLUSTERS=ops.NORMALIZE_NUMBER_CLUSTERS, MAKE_PLOTS=MAKE_PLOTS)
		print "\tFinding cluster mappings..  "
		clusterMappings = mapClusterNodes(clusterNodes, tgt, ag.sourceHandles, ops.DESCRIPTOR_HOP_SIZE_SEC, MAKE_PLOTS=MAKE_PLOTS)
		np.set_printoptions(threshold='nan')
		print "\t\tCluster Size Map (tgt->cps) =",
		for tcst, ccst in enumerate(clusterMappings[ops.NORMALIZE_CLUSTER_MAPPING]):
			print '#'+str(tcst+1)+' =', (clusterNodes['target'][tcst]['clustSize'], clusterNodes['corpus'][ccst]['clustSize']), ' ',
		print ""
		agData.commitData("cluster", checksum, tgtAssign, cpsAssign, clusterMappings, clustMinmax, clustStddev)
	else:
		print "Loading Clustering Anal for '"+str(ops.NORMALIZE_NUMBER_CLUSTERS)+"' clusters."
		tgtAssign, cpsAssign, clusterMappings, clustMinmax, clustStddev = eval(clusterData)
	actualNumberClusters = len(clusterMappings['size'])

	if MAKE_PLOTS:
		print "\tPlotting raw data"
		scatterPlot2Des(clusterMetricNames, "raw", tgt, ag.sourceHandles, '/Users/ben/Desktop/cluster/raw.png')
		#save3dData(clusterMetricNames, "raw", tgt, ag.sourceHandles, '/Users/ben/Documents/papers/2011-cmr/fig/raw3d')
	print "\t\tCluster Mapping According to", ops.NORMALIZE_CLUSTER_MAPPING, "->", clusterMappings[ops.NORMALIZE_CLUSTER_MAPPING]
	unnormalizeAccordingToClusterMappingMinMax(tgt, ag.sourceHandles, cpsAssign, clusterMappings[ops.NORMALIZE_CLUSTER_MAPPING], clustMinmax, clustStddev, clusterMetricNames, ag, method=ops.NORMALIZE_CLUSTERS_METHOD)
	# get norm coeffs now that corpus is warped to match target clusters
	ag.targetNormCoeffs, ag.sourceNormCoeffs = normalizeCoefs(analMetrics.norm, tgt, ag, analMetrics, p, ops.NORMALIZE_METHOD)
	normalizeDescriptors([tgt], analMetrics.norm, 'raw', 'norm', 'stddev', ag.targetNormCoeffs, None, None) 
	normalizeDescriptors(ag.sourceHandles, analMetrics.norm, 'raw', 'norm', 'stddev', ag.sourceNormCoeffs, None, None) 
	
	if MAKE_PLOTS: 
		scatterPlot2Des(clusterMetricNames, "raw", tgt, ag.sourceHandles, '/Users/ben/Desktop/cluster/raw-post-clust.png')
		#save3dData(clusterMetricNames, "raw", tgt, ag.sourceHandles, '/Users/ben/Documents/papers/2011-cmr/fig/norm3d')
		scatterPlot2Des(clusterMetricNames, "norm", tgt, ag.sourceHandles, '/Users/ben/Desktop/cluster/norm-'+ops.NORMALIZE_CLUSTER_MAPPING+'.png')
		#arrowPlot2Des(clusterMetricNames, "raw", "norm", ag.sourceHandles, '/Users/ben/Desktop/cluster/migration.png')
	print "\n"
	



def clusterAnal(descriptors, descriptorWeights, tgt, sourceHandles, frameSec, NORMALIZE_NUMBER_CLUSTERS=6, MAKE_PLOTS=False):
	NORMALIZE_BEFORE_CLUSTERING = True
	WEIGHT_BEFORE_CLUSTERING = True
#	USE_ALL_SEG_ENTRIES = True # false = only whole init segments, not inside
	# GET COORDINATES #
#	if USE_ALL_SEG_ENTRIES:
#	else:
	tgtCoors = np.zeros((len(tgt.segFrames), len(descriptors)))
	for segIdx, (start, end) in enumerate(tgt.segFrames):
		for desIdx, descript in enumerate(descriptors):
			tgtCoors[segIdx][desIdx] = tgt.data['raw'][descript][start]
			
	cpsCoors = np.zeros((len(sourceHandles), len(descriptors)))
	cpsNames = []
	for corpusIdx, handle in enumerate(sourceHandles):
		if handle.name not in cpsNames: cpsNames.append(handle.name)
		for desIdx, descript in enumerate(descriptors):
			cpsCoors[corpusIdx][desIdx] = handle.data['raw'][descript][(0, None)]
	# normalize
	tgtNormcoors = (tgtCoors-np.mean(tgtCoors, axis=0))/np.std(tgtCoors, axis=0)
	cpsNormcoors = (cpsCoors-np.mean(cpsCoors, axis=0))/np.std(cpsCoors, axis=0)
	# use similarity weights		
	tgtNormcoors *= descriptorWeights
	cpsNormcoors *= descriptorWeights
	# cluster it!
	tgtClusterNodesSamples = {}
	cpsClusterNodesSamples = {}
	clusterAssignments = {}
	winnerPoints = {'target': [], 'corpus': []}
	minmax = {'target': [], 'corpus': []}
	stddev = {'target': [], 'corpus': []}
	for rawCoors, normCoors, name, points in [(tgtCoors, tgtNormcoors, 'target', tgtClusterNodesSamples), (cpsCoors, cpsNormcoors, 'corpus', cpsClusterNodesSamples)]:
		if NORMALIZE_NUMBER_CLUSTERS == "target":
			clusterAssignments[name] = clusterCoordinates(normCoors, "distance", 0.5)
			NORMALIZE_NUMBER_CLUSTERS = max(clusterAssignments[name])+1
			print "\tNumber of Clusters =", NORMALIZE_NUMBER_CLUSTERS, ';', 'Number of Dimensions =', len(descriptors)
		else:
			clusterAssignments[name] = clusterCoordinates(normCoors, "maxclust", NORMALIZE_NUMBER_CLUSTERS)
		# put each coor idx into its cluster via a dict
		for idx, clusterID in enumerate(clusterAssignments[name]):
			if not points.has_key(clusterID): points[clusterID] = {'idx': []}
			points[clusterID]['idx'].append(idx)
		# find mean, std; pick winners
		for clusterId, dict in points.iteritems():
			# make a tmp array with each cluster point
			clusterCoors = np.zeros((len(dict['idx']), len(descriptors)))
			for numb, idx in enumerate(dict['idx']): clusterCoors[numb] = rawCoors[idx]
			clustData = clusterMeanStd(clusterCoors, dict['idx'])
			winnerPoints[name].append(clustData)
			minmax[name].append((clustData['mins'], clustData['maxs']))
			stddev[name].append((clustData['avg'], clustData['std']))
			del clusterCoors
		# plot it
		if MAKE_PLOTS:
			import matplotlib.pyplot as plt
			colors = ['b', 'g', 'c', 'm', 'y', 'k', 'w', 'r']
			cc = 0
			for idx, clusterID in enumerate(clusterAssignments[name]):
				thisColor = colors[clusterID%len(colors)]
				# add each coor
				plt.scatter([rawCoors[idx][0]], [rawCoors[idx][1]], c=thisColor, marker='s', s=35)
				#if name == 'corpus': plt.annotate(str(sourceHandles[idx].voiceID),  xy=(rawCoors[idx][0], rawCoors[idx][1]))
			# add average cluster node points
			for clusterID in range(NORMALIZE_NUMBER_CLUSTERS):
				thisColor = colors[(clusterID)%len(colors)]
				plt.scatter([winnerPoints[name][clusterID]['avg'][0]], [winnerPoints[name][clusterID]['avg'][1]], c=thisColor, marker='o', s=100)
			plt.savefig('/Users/ben/Desktop/cluster/'+name+'.png')
			plt.close()	
	if MAKE_PLOTS:
		from matplotlib.patches import Rectangle
		a = plt.gca()
		for collection in winnerPoints:#winnerPoints:
			for idx, point in enumerate(winnerPoints[collection]):
				# circles
				thisColor = colors[idx%len(colors)]
				plt.scatter([tgtCoors[idx][0]], [tgtCoors[idx][1]], c=thisColor, marker='s', s=35)

				rectOrigin = (point['centroid'][0]-(point['dev'][0]/2.0), point['centroid'][1]-(point['dev'][1]/2.0))
				ell = Rectangle(xy=rectOrigin, width=point['dev'][0], height=point['dev'][1]) #, angle=0
				ell.set_alpha(0.1)
				ell.set_facecolor(thisColor)
				a.add_patch(ell)
		for idx, clusterID in enumerate(clusterAssignments['target']):
			thisColor = colors[clusterID%len(colors)]
			plt.scatter([tgtCoors[idx][0]], [tgtCoors[idx][1]], c=thisColor, marker='s', s=35)
		for idx, clusterID in enumerate(clusterAssignments['corpus']):
			thisColor = colors[clusterID%len(colors)]
			plt.scatter([cpsCoors[idx][0]], [cpsCoors[idx][1]], c=thisColor, marker='o', s=35)
		plt.savefig('/Users/ben/Desktop/cluster/raw-clustered.png')
		plt.close()	
	return winnerPoints, clusterAssignments['target'], clusterAssignments['corpus'], minmax, stddev





class BreakIt(Exception): pass
## ^ used for double-loop exiting
def mapClusterNodes(clusterNodes, tgt, sourceHandles, frameSec, MAKE_PLOTS=False):
	clusterMapping = {}
	clusterMap = {} # cps nodes which corrispond to indexed tgt nodes
	numberClusters = len(clusterNodes['target'])
	normClusters = normClusterNodes(clusterNodes)
	#################
	#	BY SIZE		#
	#################
	cpsSize = tgtSize = []
	for listing, result in [(clusterNodes['target'], tgtSize), (clusterNodes['corpus'], cpsSize)]:
		for idx, dict in enumerate(listing):
			result.append( (dict['clustSize'], idx) )
		result.sort()
	clusterMap['size'] = [None]*numberClusters 
	for idx, (size, tgtClusterID) in enumerate(tgtSize): clusterMap['size'][tgtClusterID] = cpsSize[idx][1]
	###################
	#	BY DENSITY	  #
	###################
	cpsSize = tgtSize = []
	for listing, result in [(clusterNodes['target'], tgtSize), (clusterNodes['corpus'], cpsSize)]:
		for idx, dict in enumerate(listing):
			result.append( (dict['clustDensity'], idx) )
		result.sort()
	clusterMap['density'] = [None]*numberClusters 
	for idx, (size, tgtClusterID) in enumerate(tgtSize): clusterMap['density'][tgtClusterID] = cpsSize[idx][1]
	#################
	#  BY DISTANCE	#
	#################
	minimumFoundError = (sys.maxint, None) # errorAmount, permSeriesIdx
	permutations = utilities.permutations(range(numberClusters))
	for idx, perm in enumerate(permutations):
		thisError = 0
		#print idx
		try:
			for tgtIdx, cpsIdx in enumerate(perm):
				for dim in range(len(normClusters['target'][tgtIdx])):
					thisError += (normClusters['target'][tgtIdx][dim]-normClusters['corpus'][cpsIdx][dim])**2
					if thisError > minimumFoundError[0]:
						raise BreakIt # break both loops & continue if over the minimum found error..
		except BreakIt:
			pass
		if thisError < minimumFoundError[0]:
			minimumFoundError = (thisError, perm)
	clusterMap['leastSqrError']  = minimumFoundError[1]
	
	if MAKE_PLOTS:
		import matplotlib.pyplot as plt
		numbClust = len(normClusters['target'])
		colors = ['b', 'g', 'c', 'm', 'y', 'k', 'w', 'r']
		
		for method in clusterMap:
			for tgtClust, cpsClust in enumerate(clusterMap[method]):
				plt.plot([normClusters['target'][tgtClust][0] ], [ normClusters['target'][tgtClust][1] ],'g.',markersize=20.0)
				plt.scatter([normClusters['target'][tgtClust][0] ], [ normClusters['target'][tgtClust][1] ], c='b', marker='o')
				plt.annotate(str(tgtClust),  xy=normClusters['target'][tgtClust][0:2])
				plt.scatter([normClusters['corpus'][cpsClust][0] ], [ normClusters['corpus'][cpsClust][1] ], c='r', marker='o')
				plt.annotate(str(tgtClust),  xy=normClusters['corpus'][cpsClust][0:2])
			plt.savefig('/Users/ben/Desktop/cluster/mapped-'+method+'.png')
			plt.close()	
		# make csound play list
		tgtLeaves = []
		cpsLeaves = []
		for idx, dict in enumerate(clusterNodes['target']): 
			tgtLeaves.append(dict['bestIdx'])
		for idx, dict in enumerate(clusterNodes['corpus']): 
			cpsLeaves.append(dict['bestIdx'])
		timeKeeper = 0
		csoundWrites = []
		for leaf in tgtLeaves:
			start = tgt.segFrames[leaf][0]*frameSec
			duration = (tgt.segFrames[leaf][1]*frameSec)-start
			csoundWrites.append([str(timeKeeper), str(start), str(duration), tgt.filename])
			timeKeeper += duration+0.8
		for leaf in cpsLeaves:
			h = sourceHandles[leaf]
			start = h.startSec
			duration = h.segdata.getraw('effDur-seg', 0, None)*frameSec
			csoundWrites.append([str(timeKeeper), str(start), str(duration), h.filename])
			timeKeeper += duration+0.8
		# now the two together, according to cluster mapping and max segment duration..
		for leaf1, leaf2 in zip(tgtLeaves, cpsLeaves):
			h = sourceHandles[leaf2]
			start1 = tgt.segFrames[leaf1][0]*frameSec
			start2 = h.startSec
			duration1 = (tgt.segFrames[leaf1][1]*frameSec)-start1
			duration2 = h.segdata.getraw('effDur-seg', 0, None)*frameSec
			csoundWrites.append([str(timeKeeper), str(start1), str(duration1), tgt.filename])
			csoundWrites.append([str(timeKeeper), str(start2), str(duration2), h.filename])
			timeKeeper += max(duration1, duration1)+0.8
		# write the score as txt, execute csound
		scoreHandle = open('/tmp/tmp.sco', 'w')
		for (timeKeeper, start, duration, filename) in csoundWrites:
			line =[ 'i1', timeKeeper, duration, '100', '"'+filename+'"', start,'1', '0', '0', '0', '0.01', '0.01']
			scoreHandle.write(' '.join(line)+'\n')
		scoreHandle.close()
		#utilities.executeCommandQuite(['csound --output=/Users/ben/Desktop/cluster/clustered.aiff "'+os.path.join(ops.AG_PATH,'scripts','csoundRender.orc')+'" /tmp/tmp.sco > /tmp/out 2>&1'])
	return 	clusterMap	



def normClusterNodes(clusterNodes):
	dim = len(clusterNodes['target'][0]['avg'])
	numbClust = len(clusterNodes['target'])
	normClusterNodes = {}
	# add tgt points
	for name in ['target', 'corpus']:
		tmp = np.zeros((numbClust, dim))
		for idx, cluster in enumerate(clusterNodes[name]):
			tmp[idx] = cluster['bestPoint']
		mean = np.mean(tmp, axis=0)
		stdy = np.std(tmp, axis=0)
		for idx, arr in enumerate(tmp):
			tmp[idx] = (arr-mean)/stdy
		normClusterNodes[name] = tmp[:]
	return normClusterNodes





def clusterDistance(x, y, VI, method='mahalanobis'):
	if method == 'mahalanobis':
		from hcluster import mahalanobis
		return mahalanobis(x, y, VI)
	elif method == 'euclidean':
		from numpy.linalg import norm as normDist
		#print "DIST", normDist(x-y), normDist(x-(y+VI)), normDist(x-(y-VI))
		#return normDist(x-y)
		return min(normDist(x-(y+VI)), normDist(x-(y-VI)))





def plotClusterDistances(frameData, clusterMapping, distances, idx, path):
	import matplotlib.pyplot as plt
	plt.scatter([frameData[0]], [frameData[1]], marker='o') # add virtual point
	plt.annotate('tgt',  xy=(frameData[0], frameData[1]))
	
	for i, clust in enumerate(clusterMapping):
		plt.scatter([clust[0][0][0]], [clust[0][0][1]], marker='x') # add virtual point
		plt.annotate(str(distances[i]),  xy=(clust[0][0][0], clust[0][0][1]))

	plt.savefig(path+'/frame-'+str(idx)+'.png')
	plt.close()





def scatterPlot2Des(descriptors, dataType, tgt, sourceHandles, path):
	useDesc = []
	for descriptor in descriptors:
		if descriptor.find('-seg') != -1:  useDesc.append(descriptor)
	useDesc = useDesc[:2]
	# scatter plot raw data, then normalized data
	import matplotlib.pyplot as plt
	from numpy import zeros
	tgtCoors = zeros((len(useDesc), len(tgt.segFrames)))
	for segIdx, (start, end) in enumerate(tgt.segFrames):
		for desIdx, descript in enumerate(useDesc):
			tgtCoors[desIdx][segIdx] = tgt.data[dataType][descript][start]
	cpsCoors = zeros((len(useDesc), len(sourceHandles)))
	for corpusIdx, handle in enumerate(sourceHandles):
		for desIdx, descript in enumerate(useDesc):
			cpsCoors[desIdx][corpusIdx] = handle.data[dataType][descript][0]

	for array, shape, color in [(cpsCoors, 'o', 'r'), (tgtCoors, 's', 'b')]:
		plt.scatter(array[0], array[1], c=color, marker=shape) # add virtual point
	plt.savefig(path)
	plt.close()


def save3dData(descriptors, dataType, tgt, sourceHandles, path):
	useDesc = []
	for descriptor in descriptors:
		if descriptor.find('-seg') != -1:  useDesc.append(descriptor)
	useDesc = useDesc[:3]
	# scatter plot raw data, then normalized data
	from numpy import zeros
	tgtCoors = zeros((len(useDesc), len(tgt.segFrames)))
	for segIdx, (start, end) in enumerate(tgt.segFrames):
		for desIdx, descript in enumerate(useDesc):
			tgtCoors[desIdx][segIdx] = tgt.data[dataType][descript][start]
	cpsCoors = zeros((len(useDesc), len(sourceHandles)))
	for corpusIdx, handle in enumerate(sourceHandles):
		for desIdx, descript in enumerate(useDesc):
			cpsCoors[desIdx][corpusIdx] = handle.data[dataType][descript][0]
	print cpsCoors
	h1 = open(path+'-tgt.npy', 'w')
	h2 = open(path+'-cps.npy', 'w')
	np.save(h1, tgtCoors)
	np.save(h2, cpsCoors)
	h1.close()
	h2.close()
	

def arrowPlot2Des(descriptors, dataType1, dataType2, sourceHandles, path):
	import matplotlib.pyplot as plt
	useDesc = []
	for descriptor in descriptors:
		if descriptor.find('-seg') != -1: useDesc.append(descriptor)
	useDesc = useDesc[:2]
	# scatter plot raw data, then normalized data
	from numpy import zeros
	startCoors = zeros((len(useDesc), len(sourceHandles)))
	endCoors = zeros((len(useDesc), len(sourceHandles)))
	for corpusIdx, handle in enumerate(sourceHandles):
		for desIdx, descript in enumerate(useDesc):
			startCoors[desIdx][corpusIdx] = handle.data[dataType1][descript][0]
			endCoors[desIdx][corpusIdx] = handle.data[dataType2][descript][0]
		plt.arrow(startCoors[0][corpusIdx], endCoors[0][corpusIdx], startCoors[1][corpusIdx],endCoors[1][corpusIdx], width=0.01)

	for array, shape, color in [(startCoors, 'o', 'r'), (endCoors, 'o', 'r')]:
		plt.scatter(array[0], array[1], c=color, marker=shape) # add virtual point
	plt.savefig(path)
	plt.close()
#		arrowPlot2Des(clusterMetricNames, "raw", "norm", ag.sourceHandles, '/Users/ben/Desktop/cluster/migration.png')



def product(*args, **kwds):
    # product('ABCD', 'xy') --> Ax Ay Bx By Cx Cy Dx Dy
    # product(range(2), repeat=3) --> 000 001 010 011 100 101 110 111
    pools = map(tuple, args) * kwds.get('repeat', 1)
    result = [[]]
    for pool in pools:
        result = [x+[y] for x in result for y in pool]
    for prod in result:
        yield tuple(prod)