import sys
import numpy as np
import metrics
import utilities
import normalize as norm
import random
import operator

class corpusObj:
	def __init__(self, handle, stream):
		self.handle = handle
		self.stream = stream
		self.sim_accum = 0.
		self.data = {}

class BreakIt(Exception):
	pass




def printString(tgtTime, overlapNumber, passLengths):
	clockTime = (6-len(str(utilities.trunc(ag.f2s(tgtTime), 2))))*' '+str(utilities.trunc(ag.f2s(tgtTime), 2))
	firstHalfString = 'Searching: Target @ %s x %i'%(clockTime, overlapNumber+1)
	firstHalfString += ' '*(32-len(firstHalfString))
	if len(passLengths) == 1:
		secondHalf = '-   Search size: '
	else:
		secondHalf = '-   Search pass sizes: '
	for p in passLengths:
		secondHalf += '%s%i'%(' '*(6-len(str(p))), p)
	wholeString = firstHalfString+secondHalf
	return wholeString


class distanceCalculations:
	def __init__(self, descriptorData):
		self.neverMix = descriptorData.neverMix
		self.singleNumberDescriptors = descriptorData.singleNumberDescriptors
		self.searchResults = []
		self.paramScore = []
		self.segmentDensityAmpScalers = [None, None]
		if ops.SUPERIMPOSE.minOnset != None:
			self.segmentDensityAmpScalers[0] = 1/float(ops.SUPERIMPOSE.minOnset)
		else:
			self.segmentDensityAmpScalers[0] = 1/1.
		if ops.SUPERIMPOSE.maxOnset != None:
			self.segmentDensityAmpScalers[0] = 1/float(ops.SUPERIMPOSE.maxOnset)
		else:
			self.segmentDensityAmpScalers[0] = 1/8.		
		
	##############################
	def setCorpus(self, corpusHandles, validStreams):
		self.corpusObjs = [] # clear it
		for i in range(len(corpusHandles)):
			self.corpusObjs.append(corpusObj(corpusHandles[i], validStreams[i]))
	##############################
	def executeSearch(self, seach_pass_objs, tgt, tt, tl, tp, to, forceInc, log):	
		#self.stats = []
		self.searchMinMax = []
		passedLengths = [len(self.corpusObjs)] # start with initial size for printing
		for spidx, spassobj in enumerate(seach_pass_objs):
			mind = sys.maxint
			maxd = -1*sys.maxint
			if len(self.corpusObjs) < 1: return False

			if spassobj.method == 'ratio_limit':
				assert len(spassobj.timevary) == 0 and len(spassobj.seg) == 1
				newList = []
				#self.stats.append({'min': sys.maxint, 'max': -1*sys.maxint})
				for cidx, c in enumerate(self.corpusObjs):
					st, ed, dataDict = self.getStartAndEndTimeForSimCalc(c.handle, tgt, tl, tp)
					c.data.update(dataDict)
					c.data['trans'] = 0.
					c.data['availableStreams'] = c.stream 
					tgtRawVal, cpsRawVal = self.getSegmentedPointer(tgt, c.handle, spassobj.seg[0], st, ed, raw=True)
					c.ratio = cpsRawVal/float(tgtRawVal)
					# so ratio=1.1 is corpus is 110% of target
					if c.ratio > maxd: maxd = c.ratio
					if c.ratio < mind: mind = c.ratio
					#print "\t", spassobj.maxratio, c.ratio
					
					if spassobj.minratio != None and c.ratio < spassobj.minratio: continue
					if spassobj.maxratio != None and c.ratio > spassobj.maxratio: continue
					newList.append(c)
				self.corpusObjs.sort(key=operator.attrgetter('sim_accum'))
				log.write("\tsearch pass #%i - %s - %i -> %i ( %2.2f, %2.2f )\n"%(spidx+1, spassobj.method, len(self.corpusObjs), len(newList), mind, maxd))
				self.corpusObjs = newList
				#print "RATIO LIMIT", len(self.corpusObjs), len(newList)
				#print "RATIO LIMIT", len(self.corpusObjs), len(newList)
				#print "RATIO LIMIT", len(self.corpusObjs), len(newList)
				passedLengths.append(len(newList))
			else:
				#########################################################
				## search segmented and timevarying features distances ##
				#########################################################		
				min_accum = sys.maxint
				for cidx, c in enumerate(self.corpusObjs):
					c.sim_accum = 0.	
					st, ed, dataDict = self.getStartAndEndTimeForSimCalc(c.handle, tgt, tl, tp)
					c.data.update(dataDict)
					c.data['trans'] = 0.
					c.data['availableStreams'] = c.stream 
					try:
						for d in spassobj.seg:
							tgtVal, cpsVal = self.getSegmentedPointer(tgt, c.handle, d, st, ed)
							# SPECIAL EXCEPTION IF MANIPULATING POWER FOR SIMULTANEOUS LAYERING:
							if d.name == 'power-seg' and ops.RANDOMIZE_AMPLITUDE_FOR_SIM_SELECTION and self.segmentDensityAmpScalers[0]!= self.segmentDensityAmpScalers[1]: # a bool
								tgtVal *= random.uniform(self.segmentDensityAmpScalers[0], self.segmentDensityAmpScalers[1])							
							dist = ((tgtVal-cpsVal)**2)*(d.weight*c.handle.scaleDistance)
							c.sim_accum += dist
							if c.sim_accum > min_accum and not spassobj.complete_results: raise BreakIt
							
						for d in spassobj.timevary:
							tgtVals, cpsVals = self.getTimeVaryingFeaturePointer(tgt, c.handle, d, st, ed)
							dist = timeVaryingDistance(tgtVals, cpsVals, dist=d.distance, envelopeMask=c.handle.envelopeMask, energyWeight=d.energyWeight, energies=c.handle.data['raw']['power'], peaks=(tgt.segdata.getraw('peakTime-seg', 0, None), c.handle.segdata.getraw('peakTime-seg', 0, None)))
							c.sim_accum += dist*d.weight*c.handle.scaleDistance
							if c.sim_accum > min_accum and not spassobj.complete_results: raise BreakIt
						if c.sim_accum < min_accum: min_accum = c.sim_accum
					except BreakIt: pass
				# sort by accum distance
				self.corpusObjs.sort(key=operator.attrgetter('sim_accum'))
				mind = self.corpusObjs[0].sim_accum
				maxd = self.corpusObjs[-1].sim_accum
				
				lastPass = spidx == len(seach_pass_objs)-1				
				# clip corpus list to reflect search results and scope
				if spassobj.method == 'closest':
					newList = [self.corpusObjs[0]]
				elif spassobj.method == 'farthest':
					newList = [self.corpusObjs[-1]]
				else:
					assert spassobj.percent != None
					numb_entries = max(2, int(len(self.corpusObjs)*(spassobj.percent/100.))) # two!
					if spassobj.method == 'closest_percent':
						newList = self.corpusObjs[:numb_entries]
					elif spassobj.method == 'farthest_percent':
						newList = self.corpusObjs[numb_entries:]
				
				### if this is the last pass
				if lastPass and len(newList) > 1: # make a random choice
					newList = [random.choice(newList)]
				log.write("\tpass #%i - %s - %i -> %i ( %2.2f %f2.2f )\n"%(spidx+1, spassobj.method, len(self.corpusObjs), len(newList), mind, maxd))
				self.corpusObjs = newList
				
				if not lastPass:
					passedLengths.append(len(newList))
				
			p.post(printString(tt, to, passedLengths), inc=True, forceInc=forceInc)
		# done with loop
		assert len(self.corpusObjs) == 1
		self.searchResults.append( (self.corpusObjs[0].handle, self.corpusObjs[0].data) )
		self.searchMinMax.append([mind, maxd])
		return True



	##############################
		# hacky + experimental
		# hacky + experimental
		# hacky + experimental
	def paramInterpolateScore(self, PARAM_DIR):
		import os
		PARAM_INTERP_NUMB_PARAMS = False
		parameterValues = {}
		uninterpParameterValues = {}
		# figure out number of parameters...
		testData = self.corpusObjs[0].handle.extraMappedData
		for fidx, field in enumerate(testData):
			if field == 'params':
				PARAM_INTERP_NUMB_PARAMS = int(testData[fidx+1])
				mappedDataParamSlice = slice(fidx+2, fidx+2+PARAM_INTERP_NUMB_PARAMS)
		print "\n\nINTERP PARAMS -- found", PARAM_INTERP_NUMB_PARAMS, "parameters..."
		paramFoundFoundFloat = 100000.
		# initialize dict holding values for each of the parameters for each frame of the target
		for pidx in range(PARAM_INTERP_NUMB_PARAMS):
			parameterValues[pidx] = np.zeros(tgt.lengthInFrames)
			uninterpParameterValues[pidx] = np.zeros(tgt.lengthInFrames)
			parameterValues[pidx] += paramFoundFoundFloat
			uninterpParameterValues[pidx] += paramFoundFoundFloat
		# now look through search results one by one...		
		for dict in self.searchResults:
			if dict['pass'] != 'DESCRIPTORS': continue
			params = []
			weights = []
			for idx in range(ops.SEARCH_CORPUS_PARAM_INTERPOLATE_POINTS):
				try:
					handle = self.corpusObjs[dict['list'][idx][1]].handle
					weights.append(dict['list'][idx][0])
					params.append( [float(x) for x in handle.extraMappedData[mappedDataParamSlice]] )
				except IndexError: continue
			interpolatedParameters = np.average(np.array(params), weights=weights, axis=0)
			thisEntry = [dict['time'], dict['length']]
			#parameterTimeArraySlice = slice(dict['time'], dict['time']+dict['length'])
			for pidx, parameter in enumerate(interpolatedParameters):
				parameterValues[pidx][dict['time']] = parameter
			for pidx, parameter in enumerate(params[0]):
				uninterpParameterValues[pidx][dict['time']] = parameter
		# make sure parameters don't give zeros, but stay in the correct ranges
		for pidx in range(PARAM_INTERP_NUMB_PARAMS):
			lastVal = 0.
			for vidx in range(tgt.lengthInFrames):
				if parameterValues[pidx][vidx] == paramFoundFoundFloat:
					parameterValues[pidx][vidx] = lastVal
				else:
					lastVal = parameterValues[pidx][vidx]
		# write parameter f-table files
		targetMaxPower = np.max(tgt.data['orig']['power'])
		fPower = open(os.path.join(PARAM_DIR, 'paramAmp.txt'), 'w')
		for time in range(tgt.lengthInFrames):
			#if tgt.thresholdTest(time):
			fPower.write( str(  tgt.data['orig']['power'][time]/targetMaxPower  )+'\n' )
			#else:
			#	fPower.write( '0.0\n' ) # didn't pass !
# make sure target passes threshold test first...
#			if tgt.thresholdTest(time):
#				fPower.write( str( (tgt.data['orig']['power'][time]/targetMaxPower)**0.5 )+'\n' )
#			else:
#				fPower.write( '0.0\n' ) # didn't pass !
		fPower.close()
		fcentroid = open(os.path.join(PARAM_DIR, 'paramCentroid.txt'), 'w')
		for time in range(tgt.lengthInFrames):
			fcentroid.write( str( tgt.data['orig']['centroid'][time] )+'\n' )
# make sure target passes threshold test first...
#			if tgt.thresholdTest(time):
#				fPower.write( str( (tgt.data['orig']['power'][time]/targetMaxPower)**0.5 )+'\n' )
#			else:
#				fPower.write( '0.0\n' ) # didn't pass !
		fcentroid.close()

		# PLOT THAT SHIT
		#allLabels = ['power']
		#allParams = np.append(parameterValues, tgt.data['orig']['power'])
		#print allParams
		#print allParams.shape
#		allLabels = []
#		for pidx in range(PARAM_INTERP_NUMB_PARAMS):
#			allLabels.append('param%i'%pidx)
#		utilities.plot(os.path.join(PARAM_DIR, 'paramPlot.png'), parameterValues.values(), allLabels, segs=False, normalize=True, height=15, width=45, extraText=[])



		for pidx in range(PARAM_INTERP_NUMB_PARAMS):
			fparam = open(os.path.join(PARAM_DIR, 'param'+str(pidx)+'.txt'), 'w')
#			fparamUn = open(os.path.join(PARAM_DIR, 'paramUninterp'+str(pidx)+'.txt'), 'w')
			for param in parameterValues[pidx]:
				fparam.write( str(param)+'\n' )
#			for param in uninterpParameterValues[pidx]:
#				fparamUn.write( str(param)+'\n' )
			fparam.close()
#			fparamUn.close()

		
		

	##############################
	def peakMatcher(self, chandle, tgt, length, st, ed, peakTime, whichDescriptors, thisPassSegDescript):
		peakNumbers = []
		descriptDelete = []
		for didx, (descript,dweight,foundin) in enumerate(thisPassSegDescript):
			if whichDescriptors not in foundin: continue # make sure descript matches this pass, first pass, etc.
			if descript.find('peakFrq') != -1:
				peakNumbers.append(descript.split('-')[1])
				descriptDelete.append(didx)
		# then there are no peak descriptors, so don't do anything!!
		if len(peakNumbers) == 0: return 0., descriptDelete, 0.
		# gather peaks and amp into a single array!
		tgtpeaks = []
		cpspeaks = []
		tgtamps = []
		cpsamps = []
		# instead of taking segmented averages, lets take the peaks at max power time...
		tgtMaxPowerTime = peakTime
		cpsMaxPowerTime = int(chandle.data['raw']['peakTime-seg'][(0, None)])
		MATCH_PEAKS_METHOD = 'peakTime'
		if MATCH_PEAKS_METHOD == 'peakTime':
			for n in peakNumbers:
				# must be raw in order to get transposition correlation to work!
				if tgt.data['raw']['peakAmp-'+n][tgtMaxPowerTime] > ops.MATCH_PEAKS_AMP_THRESH:
					tgtpeaks.append(tgt.data['raw']['peakFrq-'+n][tgtMaxPowerTime])
					tgtamps.append(tgt.data['raw']['peakAmp-'+n][tgtMaxPowerTime])
				if chandle.data['raw']['peakAmp-'+n][cpsMaxPowerTime] > ops.MATCH_PEAKS_AMP_THRESH:
					cpspeaks.append(chandle.data['raw']['peakFrq-'+n][cpsMaxPowerTime])
					cpsamps.append(chandle.data['raw']['peakAmp-'+n][cpsMaxPowerTime])
		elif MATCH_PEAKS_METHOD == 'averages':
			for n in peakNumbers:
				# must be` raw in order to get transposition correlation to work!
				tgtFrq = tgt.data['raw']['peakFrq-'+n][(st, ed)]
				tgtAmp = tgt.data['raw']['peakAmp-'+n][(st, ed)]

				cpsFrq = chandle.data['raw']['peakFrq-'+n+'-seg'][(0, None)]
				cpsAmp = chandle.data['raw']['peakAmp-'+n+'-seg'][(0, None)]

				if tgtAmp > ops.MATCH_PEAKS_AMP_THRESH:
					tgtpeaks.append(tgtFrq)
					tgtamps.append(tgtAmp)
				if cpsAmp > ops.MATCH_PEAKS_AMP_THRESH:
					cpspeaks.append(cpsFrq)
					cpsamps.append(cpsAmp)

		minPeaksFound = min(len(tgtpeaks), len(cpspeaks))
		if minPeaksFound == 0:  return sys.maxint, descriptDelete, 0.
		
		tgtpeaks = np.array(tgtpeaks[:minPeaksFound])
		cpspeaks = np.array(cpspeaks[:minPeaksFound])
		#print tgtpeaks, '->',  cpspeaks
		#print
		if ops.MATCH_PEAKS_CHROMA_ONLY: cpspeaks = cpspeaks % 12.
		# transpose target up and down to get pitch variation...
		transIntervals = np.linspace(-ops.MATCH_PEAKS_TRANS_UP_AND_DOWN, ops.MATCH_PEAKS_TRANS_UP_AND_DOWN, int(ops.MATCH_PEAKS_TRANS_UP_AND_DOWN/(ops.MATCH_PEAKS_TRANS_ITERATE_INTERVAL/2.))+1)
		tgttransforms = []
		for inty in transIntervals:
			tgttransforms.append((tgtpeaks-inty, inty))
		minimumFoundError = [sys.maxint, None, None, None, None] # errorAmount, permSeriesIdx, Transpose, peak mapping
		for tgtpeaksTrans, transp in tgttransforms:
			permutations = utilities.permutations(range(minPeaksFound)) # every time to reset the generator object
			#print "peaks", tgtpeaksTrans, transp
			if ops.MATCH_PEAKS_CHROMA_ONLY: tgtpeaksTrans = tgtpeaksTrans % 12.
			# least square distance
			for idx, perm in enumerate(permutations):
				#print "\tpermutations", perm, 'trans', transp
				thisError = 0
				try:
					for tgtIdx, cpsIdx in enumerate(perm):
						errorAdd = ((tgtpeaksTrans[tgtIdx]-cpspeaks[cpsIdx])**2)
						if ops.MATCH_PEAKS_USE_PARTIAL_AMPLITUDE: errorAdd *= cpsamps[cpsIdx]
						thisError += ((tgtpeaksTrans[tgtIdx]-cpspeaks[cpsIdx])**2)
						#print "\t\t", tgtpeaksTrans[tgtIdx], cpspeaks[cpsIdx]
						#print "\t\t", tgtpeaksTrans, cpspeaks
						if thisError > minimumFoundError[0]:
								raise BreakIt # break both loops & continue if over the minimum found error..
				except BreakIt:
					pass
				thisError /= float(len(peakNumbers)) # divde by number of peaks..
				if thisError < minimumFoundError[0]:
					minimumFoundError = [thisError, perm, transp, tgtpeaksTrans, cpspeaks]
#		print
#		print
#		print minimumFoundError
#		print
#		print
		return minimumFoundError[0], descriptDelete, minimumFoundError[2]


# peakMatch
#	if 'peakFrq-0-seg' in tgt.data['raw'].keys():
#		for descript,dweight,foundin in self.segDescript:
#			if whichDescriptors not in foundin: continue # make sure descript matches this pass, first pass, etc.
#			for (descript,dweight,foundin) in self.segDescript:
#				print descript,dweight,foundin
#			sys.exit()
#		tgtVal, cpsVal = self.getSegmentedPointer(tgt, c.handle, descript, st, ed)
#		

	##############################
#	def retainpercent(self, percent):
#		tmp = []
#		outputList = self.searchResults[-1]['list'] # take last query
#		numberResults = int(len(outputList)*(percent/100.))
#		if numberResults == 0: numberResults = 1
#		for kidx in range(numberResults):
#			tmp.append(self.corpusObjs[outputList[kidx][1]]) # orig cps idx
#		self.corpusObjs = tmp
#	##############################
#	def returnSampleHandle(self, match, matchPercent):
#		outputList = self.searchResults[-1]['list'] # take last query
#		if ops.MATCH == "best":
#			returnIdx = 0
#		elif ops.MATCH == "randomInTopPercent":
#			randomPercent = random.uniform(0, (ops.MATCH_PERCENT/100.))
#			returnIdx = int(len(outputList)*randomPercent)
#		elif ops.MATCH == "percentExclude":
#			returnIdx = int(len(outputList)*(ops.MATCH_PERCENT/100.))
#		elif ops.MATCH == "worst":
#			returnIdx = len(outputList)-1 # last one
#		winningHandle = self.corpusObjs[outputList[returnIdx][1]].handle
#		return winningHandle, [winningHandle, {}, outputList[returnIdx][2]]
	##############################
	def close(self):
		pass
	##############################
#	def writeXml(self, xmlObj):
#		outputFrameCounters = {}
#		for outputDict in self.searchResults:
#			if outputDict['pass'] != 'DESCRIPTORS': continue
#			print outputDict['pass']
#			if not outputFrameCounters.has_key(outputDict['time']): outputFrameCounters[outputDict['time']]=0
#			else: outputFrameCounters[outputDict['time']] += 1
#			
#			hashTag = str(outputDict['time'])+'-'+str(outputFrameCounters[outputDict['time']])
#			print hashTag
#			
#			for (weight, cidx, dataDict) in outputDict['list']:
#				print (weight, cidx)
#				print self.corpusObjs[cidx].handle.name, "->", weight
#				
##				xmlObj.commitData("cluster", checksum, tgtAssign, cpsAssign, clusterMappings)
	##############################
	def getTimeVaryingFeaturePointer(self, tgt, srcHandle, d, st, ed):
		tgtArray = tgt.data['norm'][d.name][st:ed]
		if ops.SUPERIMPOSE.calcMethod in [None, "subtract"]:
			srcArray = srcHandle.data['norm'][d.name][0:ed-st]
		elif ops.SUPERIMPOSE.calcMethod == "mixture":
			if len(ag.activeSegments) == 0 or d.name not in tgt.data['mix-raw'].keys():
				srcArray = srcHandle.data['norm'][d.name][0:ed-st]
			else:
				mixDuration = ed-st
				if ed > tgt.lengthInFrames: mixDuration = tgt.lengthInFrames-st
				if mixDuration > srcHandle.lengthInFrames:
					mixDuration = srcHandle.lengthInFrames # if too long for corpus item
					ed = st+mixDuration
				rawOutput =  metrics.mixTwoTimeVaryingFeatures(d.name, st, mixDuration, tgt.data['mix-raw'], srcHandle.data['raw'])
				srcArray = norm.normDataArray(d.name, rawOutput, 'stddev', ag.sourceNormCoeffs)
		return tgtArray, srcArray
	##############################
	def getSegmentedPointer(self, tgt, srcHandle, d, st, ed, raw=False):
		cpsEffDur =  srcHandle.segdata.getraw('effDur-seg', 0, None)
		targetRawOutput =  tgt.segdata.getraw(d.name, st, ed)
		
		#if d.name == 'mfcc1-seg':
		#	print st, ed, targetRawOutput
		targetVal = tgt.segdata.getnorm(d.name, st, ed)
		
		if ops.SUPERIMPOSE.calcMethod in [None, "subtract"]:
			rawSegOutput = srcHandle.segdata.getraw(d.name, 0, cpsEffDur)
			#srcHandle.data['raw'][d.name][(0, None)] # seg data
			sourceVal = srcHandle.segdata.getnorm(d.name, 0, cpsEffDur)
			#srcHandle.data['norm'][d.name][(0, None)] # seg data
		elif ops.SUPERIMPOSE.calcMethod == "mixture":
			if len(ag.activeSegments) == 0 or d.name not in tgt.data['mix-raw'].keys():
				rawSegOutput = srcHandle.segdata.getraw(d.name, 0, cpsEffDur)
				sourceVal = srcHandle.segdata.getnorm(d.name, 0, cpsEffDur)
			else:
				mixDuration = ed-st
				if ed > tgt.lengthInFrames: mixDuration = tgt.lengthInFrames-st
				if mixDuration > srcHandle.lengthInFrames:
					mixDuration = srcHandle.lengthInFrames # if too long for corpus item
					ed = st+mixDuration
				#print "SIMCALC", st, ed, mixDuration, tgt.lengthInFrames
#				print d.name, d.parents
#				print self.singleNumberDescriptors

				assert len(d.parents) == 1
				rawTimeVarying =  metrics.mixTwoTimeVaryingFeatures(d.parents[0], st, mixDuration, tgt.data['mix-raw'], srcHandle.data['raw'])
				rawPowerSum = metrics.mixTwoTimeVaryingFeatures('power', st, mixDuration, tgt.data['mix-raw'], srcHandle.data['raw'])
				#print "\n\n", rawTimeVarying, rawPowerSum, "\n\n"
				try: rawSegOutput = np.average(rawTimeVarying, weights=rawPowerSum)
				except ZeroDivisionError:
					rawSegOutput = 0
					pass
				sourceVal = norm.normDataArray(d.name, [rawSegOutput], 'stddev', ag.sourceNormCoeffs)[0]
		if raw:
			return targetRawOutput, rawSegOutput
		return targetVal, sourceVal
	##############################
	def getStartAndEndTimeForSimCalc(self, srcHandle, tgt, targetSegmentLength, tgtPeakTime):
		# get start and end time for this similarity calculation
		if ops.SUPERIMPOSE.simCalcDur == "least":
			lengthFromTimeInFrames = min([targetSegmentLength, srcHandle.segdata.getraw('effDur-seg', 0, None)])
		elif ops.SUPERIMPOSE.simCalcDur == 'most':
			lengthFromTimeInFrames = max([targetSegmentLength, srcHandle.segdata.getraw('effDur-seg', 0, None)])
		elif ops.SUPERIMPOSE.simCalcDur == "corpusEffDur":
			#srcHandle.segdata.getraw('effDur-seg', 0, None)
			lengthFromTimeInFrames = srcHandle.segdata.getraw('effDur-seg', 0, None)
		elif ops.SUPERIMPOSE.simCalcDur == "targetEffDur":
			lengthFromTimeInFrames = targetSegmentLength
		elif ops.SUPERIMPOSE.simCalcDur == "corpusDur":	
			lengthFromTimeInFrames = srcHandle.lengthInFrames
		st = tgt.timeInFrames
		if ops.SIM_CALC_PEAK_ALIGN == 'align':
			peakCorrect = int(tgtPeakTime-srcHandle.data['raw']['peakTime-seg'][(0, None)]) # align peaks
			st += peakCorrect
		# ensure that target idx >= 0, end isn't longer than the target data
		st = max(st, 0)
		ed = min(st+lengthFromTimeInFrames, tgt.lengthInFrames)

		# store other stuff in a data dict
		data = {}
		if ops.CONCATENATE_PEAK_ALIGN == 'none':
			srcSt = tgt.timeInFrames # no correction
		elif ops.CONCATENATE_PEAK_ALIGN == 'align':
			srcSt = st # align peaks in output files
			if srcSt < 0 : srcSt = 0 # ensure that source idx > 0
		data['corpusDur'] = srcHandle.lengthInFrames
		data['targetDur'] = targetSegmentLength
		data['minDur'] = min(srcHandle.lengthInFrames, targetSegmentLength )
		data['corpusEffDur'] = srcHandle.segdata.getraw('effDur-seg', 0, None)
		data['simCalcDur'] = ed-st
		data['simCalcStEnd'] = st, ed
		data['segmentSt'] = int(srcSt)
		data['segmentEnd'] = int(srcSt)+srcHandle.lengthInFrames
		return st, ed, data
	
#tgt 2  |  |
#src 3  |   |



def interpArray(array, desiredSize, interpMask=None):
	x = np.arange(array.size)
	if interpMask == None:
		interpMask = np.linspace(0, len(array)-1, desiredSize)
	return np.interp(interpMask, x, array)




def timeVaryingDistance(array1, array2, dist=None, envelopeMask=None, energyWeight=False, energies=None, peaks=None):
	# slices are NOT copies in numpy!!!!
	l = min([len(array1), len(array2)])
	if dist == 'pearson':
		return pearsonCorr(array1[0:l], array2[0:l])
	elif dist == 'euclidean':
		distances = np.power((array1[0:l]-array2[0:l]), 2)

		
		if energyWeight: distances *= energies[0:l]
			
		#print "NEW", distances
		return np.sum(distances)/float(l) # divided by the length
	elif dist == 'kullback':
		from scipy.stats import norm 
		a1 = norm.cdf(array1[0:l])
		return (a1 * np.log2(a1/norm.cdf(array2[0:l]))).sum() / float(l) # should i divide by length here?
	elif dist.find('fixedSize') != -1: # written as fixedSize-2, fixedSize-4, etc.
		''' after philippe esling... takes normalised continuously valued features and creates
		a interpolated fixed size version of fixedSizes[0] peak and fixedSizes[1] points'''
		userSize = int(dist.split('-')[1])
		fixedSizes = (1, userSize-1) # points pre-atk, after-atk ; additional point added for peak itself
		if peaks[0] != None and peaks[1] != None and peaks != None: # interpolate arrays according to peak times...
			lenMask = fixedSizes[0]+fixedSizes[1]+1
			interpMask1 = np.zeros(lenMask)
			interpMask2 = np.zeros(lenMask)
			interpMask1[fixedSizes[0]] = peaks[0] # assign peak time idx
			interpMask2[fixedSizes[0]] = peaks[1] # assign peak time idx
			for idx in range(fixedSizes[0]): # pre peak time
				interpMask1[idx] = peaks[0]/float(fixedSizes[0]+1)
				interpMask2[idx] = peaks[1]/float(fixedSizes[0]+1)
			for idx in range(fixedSizes[0]+1, lenMask): # post peak time
				scalarBetweenPeakAndEnd = ((idx-1)/float(fixedSizes[1]+1))
				interpMask1[idx] = scalarBetweenPeakAndEnd*(l-peaks[0])+peaks[0]
				interpMask2[idx] = scalarBetweenPeakAndEnd*(l-peaks[1])+peaks[1]
			array1Fixed = interpArray(array1, lenMask, interpMask=interpMask1)
			array2Fixed = interpArray(array2, lenMask, interpMask=interpMask2)
			#print "#1", array1Fixed
			#print "#2", array2Fixed
		else: # no peak data, flat interpolation
			lenMask = fixedSizes[0]+fixedSizes[1]+1
			array1Fixed = interpArray(array1, lenMask)
			array2Fixed = interpArray(array2, lenMask)
		distances = np.power((array1Fixed-array2Fixed), 2)
		return np.sum(distances)/float(l) # divided by the length
	
	
	else:
		utilities.error('function', "unknown method for timeVaryingDistance:", dist)




def energyWeight(featureArray, energyArray):
	# with Diemo Schwartz - (metric[i]/sumWeights)*weight[i]
	weighted = (featureArray/energyArray)*np.sum(energyArray)
	weighted = metrics.removeNans(weighted, 0) # remove /0
	return weighted


def pearsonCorr(x,y):
	'''pearson correlation, return between -1,1
	-1 if inversely corrlated, 0 if no correlation
	+1 if identically correlated'''
	from itertools import imap
	n = len(x)
	sum_x = sum(x)
	sum_y = sum(y)
	sum_x_sq = sum(map(lambda x: pow(x, 2), x))
	sum_y_sq = sum(map(lambda x: pow(x, 2), y))
	psum = sum(imap(lambda x, y: x * y, x, y))
	num = psum - (sum_x * sum_y/n)
	den = pow((sum_x_sq - pow(sum_x, 2) / n) * (sum_y_sq - pow(sum_y, 2) / n), 0.5)
	if den == 0: return 0
	return num / den
	


